package com.mcaps.core.workflows;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import javax.activation.MimetypesFileTypeMap;
import javax.jcr.Node;
import javax.jcr.RepositoryException;
import javax.jcr.Session;

import org.apache.sling.api.resource.Resource;
import org.apache.sling.api.resource.ResourceResolver;
import org.osgi.framework.Constants;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.adobe.granite.workflow.WorkflowException;
import com.adobe.granite.workflow.WorkflowSession;
import com.adobe.granite.workflow.exec.WorkItem;
import com.adobe.granite.workflow.exec.WorkflowProcess;
import com.adobe.granite.workflow.metadata.MetaDataMap;
import com.day.cq.dam.api.Asset;
import com.day.cq.dam.api.AssetManager;
import com.day.cq.search.QueryBuilder;

@Component(property = {
    Constants.SERVICE_DESCRIPTION + "=Extracts a SCORM wrapped course package and puts it somewhere",
    Constants.SERVICE_VENDOR + "=Slalom LLC",
    "process.label" + "=001 Extract SCORM Course Package"
})
public class ScormExtractor implements WorkflowProcess {
    private static final Logger log = LoggerFactory.getLogger(ScormExtractor.class);

    ResourceResolver resourceResolver;
    Session session;

    @Reference
    QueryBuilder queryBuilder;

    @Override
    public void execute(WorkItem wi, WorkflowSession ws, MetaDataMap mdm) throws WorkflowException {
        resourceResolver = ws.adaptTo(ResourceResolver.class);
        session = ws.adaptTo(Session.class);

        /* Created for manually triggered workflows */
        String payloadPath = wi.getContentPath().replace("/jcr:content/renditions/original","");
        String[] params = mdm.get("PROCESS_ARGS", "string").toString().split(",");
        String[] fileExtensions = params; // Valid extensions com from the 'Arguments' field in the workflows 'Process' tab
       
        unzipFile(payloadPath, fileExtensions);
    }

    private void unzipFile(String filePath, String[] validFileExtensions) {//, String outputFolder) {
        String nodePath = filePath.replaceAll("(?:.[^/]+).zip(/jcr:content/renditions/original)?",""); // remove filename to get AEM pathing for original zip file
        Pattern p = Pattern.compile("(?:.[^/]+).("+String.join("|", validFileExtensions)+")(/jcr:content/renditions/original)?");   // pattern to search for filename only

        Resource zipfile = resourceResolver.getResource(filePath);
        AssetManager am = resourceResolver.adaptTo(AssetManager.class);

        Asset a = zipfile.adaptTo(Asset.class);
        InputStream is = a.getOriginal().getStream();

        //NOTE: now working in initial directory level, need to extract individual assets from zip
        try (ZipInputStream zipIn = new ZipInputStream(is)) {
            ZipEntry entry = zipIn.getNextEntry();
            while (entry != null) {
                Matcher m = p.matcher(entry.getName()); // set up matcher to parse original filename
                String matchedString = m.find()?m.group(0):"no match"; // extract filename w/o pathing
                String extractedFilename = matchedString.replace("/", ""); // filename w/o leading '/'

                String[] fileName = entry.getName().split("\\."); // filename split to get extension
                String fileExtension = fileName.length >= 2 ? fileName[fileName.length-1]:"none"; // file extension
                boolean extensionIsValid = Arrays.stream(validFileExtensions).anyMatch(fileExtension::equals); // file extension check

                // Extract files
                if (!entry.isDirectory() && extensionIsValid) {
                    File f = new File(extractedFilename); //create temporary file
                    try (FileOutputStream fos = new FileOutputStream(f)) {
                        byte[] buffer = new byte[1024];
                        int length;
                        while ((length = zipIn.read(buffer)) >= 0) {
                            fos.write(buffer, 0, length);
                        }

                        log.warn("SCORM WORKFLOW: saving to AEM: " + extractedFilename);
                        InputStream fis = new FileInputStream(f);
                        String aemFilePath = nodePath + matchedString; // AEM DAM path with filename

                        MimetypesFileTypeMap fileTypeMap = new MimetypesFileTypeMap();
                        String mimeType = fileTypeMap.getContentType(f.getName()); // get MIME type
                        
                        am.createAsset(aemFilePath, fis, mimeType, true); // save to AEM DAM
                        f.delete(); //delete temporary file once it has been saved to AEM.
                        setAssetMetadata(aemFilePath, a);
                    }
                }
                
                zipIn.closeEntry();
                entry = zipIn.getNextEntry();
            }
        } catch (FileNotFoundException e) {
            log.error(e.getMessage());
            e.printStackTrace();
        } catch (IOException e) {
            log.error(e.getMessage());
            e.printStackTrace();
        }
    }

    private void setAssetMetadata(String filePath, Asset parentAsset) {
        Resource r = resourceResolver.getResource(filePath);
        Node assetNode = r.adaptTo(Node.class);

        String parentLMSID = parentAsset.getMetadataValue("mcaps:lmsId");
        String skills = parentAsset.getMetadataValue("mcaps:skills");
        String[] parentSkills = skills.replace(", ", ",").split(",");
        String parentCategories = parentAsset.getMetadataValue("mcaps:category");
        String parentCourseUrl = parentAsset.getMetadataValue("mcaps:courseUrl");

        try {
                Node metadataNode = assetNode.getNode("jcr:content/metadata");
                metadataNode.setProperty("mcaps:lmsId", parentLMSID);
                metadataNode.setProperty("mcaps:skills", parentSkills);
                metadataNode.setProperty("mcaps:categories", parentCategories);
                metadataNode.setProperty("mcaps:courseUrl", parentCourseUrl);
                session.save();
        } catch (RepositoryException e) {
            e.printStackTrace();
        }
    }
}